@@ -1113,6 +1113,75 @@ void TestUses::useNamespace()
 }
 
 
+void TestUses::useGroupedNamespace()
+{
+    //                         0         1         2         3         4         5         6         7
+    //                         01234567890123456789012345678901234567890123456789012345678901234567890123456789
+    TopDUContext* top = parse("<?php\n"
+                              "namespace Foo\\Bar { class A{} }\n"
+                              "namespace Foo\\Bar { class D{} }\n"
+                              "namespace {\n"
+                              "use Foo\\Bar\\{A,D};\n"
+                              "}\n", DumpNone);
+    QVERIFY(top);
+    DUChainReleaser releaseTop(top);
+    DUChainWriteLocker lock;
+
+    Declaration* dec;
+    dec = top->findDeclarations(QualifiedIdentifier(QStringLiteral("foo"))).first();
+    QCOMPARE(dec->kind(), Declaration::Namespace);
+    compareUses(dec, QList<RangeInRevision>() << RangeInRevision(2, 10, 2, 13)
+                                          << RangeInRevision(4, 4, 4, 7) );
+
+    dec = top->findDeclarations(QualifiedIdentifier(QStringLiteral("foo::bar"))).first();
+    QCOMPARE(dec->kind(), Declaration::Namespace);
+    compareUses(dec, RangeInRevision(4, 8, 4, 11));
+
+    dec = top->findDeclarations(QualifiedIdentifier(QStringLiteral("foo::bar::a"))).first();
+    compareUses(dec, RangeInRevision(4, 13, 4, 14));
+
+    dec = top->findDeclarations(QualifiedIdentifier(QStringLiteral("foo::bar::d"))).first();
+    compareUses(dec, RangeInRevision(4, 15, 4, 16));
+}
+
+
+void TestUses::useGroupedPartialNamespace()
+{
+    //                         0         1         2         3         4         5         6         7
+    //                         01234567890123456789012345678901234567890123456789012345678901234567890123456789
+    TopDUContext* top = parse("<?php\n"
+                              "namespace Foo\\Bar { class A{} }\n"
+                              "namespace Foo\\Baz { class D{} }\n"
+                              "namespace {\n"
+                              "use Foo\\{Bar\\A,Baz\\D};\n"
+                              "}\n", DumpNone);
+    QVERIFY(top);
+    DUChainReleaser releaseTop(top);
+    DUChainWriteLocker lock;
+
+    Declaration* dec;
+    dec = top->findDeclarations(QualifiedIdentifier(QStringLiteral("foo"))).first();
+    QCOMPARE(dec->kind(), Declaration::Namespace);
+    compareUses(dec, QList<RangeInRevision>() << RangeInRevision(2, 10, 2, 13)
+                                          << RangeInRevision(4, 4, 4, 7) );
+
+    // TODO: Remaining implementation for https://bugs.kde.org/show_bug.cgi?id=411588
+    // dec = top->findDeclarations(QualifiedIdentifier(QStringLiteral("foo::bar"))).first();
+    // QCOMPARE(dec->kind(), Declaration::Namespace);
+    // compareUses(dec, RangeInRevision(4, 9, 4, 12));
+
+    // dec = top->findDeclarations(QualifiedIdentifier(QStringLiteral("foo::baz"))).first();
+    // QCOMPARE(dec->kind(), Declaration::Namespace);
+    // compareUses(dec, RangeInRevision(4, 15, 4, 18));
+
+    dec = top->findDeclarations(QualifiedIdentifier(QStringLiteral("foo::bar::a"))).first();
+    compareUses(dec, RangeInRevision(4, 13, 4, 14));
+
+    dec = top->findDeclarations(QualifiedIdentifier(QStringLiteral("foo::baz::d"))).first();
+    compareUses(dec, RangeInRevision(4, 19, 4, 20));
+}
+
+
 void TestUses::useNamespaceFunctionConst()
 {
     //                         0         1         2         3         4         5         6         7
@@ -1811,5 +1880,112 @@ void TestUses::propertyType() {
     compareUses(a, QList<RangeInRevision>() << RangeInRevision(0, 20, 0, 21));
 }
 
+void TestUses::functionParameterUnionType() {
+    //                 0         1         2         3         4         5         6
+    //                 0123456789012345678901234567890123456789012345678901234567890
+    QByteArray method("<? class A {} class B {} class C {} function foo(A|B|C $a) {}");
+    TopDUContext *top = parse(method, DumpNone);
+    DUChainReleaser releaseTop(top);
+    DUChainWriteLocker lock(DUChain::lock());
+
+    Declaration *a = top->localDeclarations().at(0);
+    QCOMPARE(a->identifier().toString(), QString("a"));
+    compareUses(a, QList<RangeInRevision>() << RangeInRevision(0, 49, 0, 50));
+
+    Declaration *b = top->localDeclarations().at(1);
+    QCOMPARE(b->identifier().toString(), QString("b"));
+    compareUses(b, QList<RangeInRevision>() << RangeInRevision(0, 51, 0, 52));
+
+    Declaration *c = top->localDeclarations().at(2);
+    QCOMPARE(c->identifier().toString(), QString("c"));
+    compareUses(c, QList<RangeInRevision>() << RangeInRevision(0, 53, 0, 54));
 }
 
+void TestUses::functionReturnUnionType() {
+    //                 0         1         2         3         4         5         6
+    //                 0123456789012345678901234567890123456789012345678901234567890
+    QByteArray method("<? class A {} class B {} class C {} function foo(): A|B|C {}");
+    TopDUContext *top = parse(method, DumpNone);
+    DUChainReleaser releaseTop(top);
+    DUChainWriteLocker lock(DUChain::lock());
+
+    Declaration *a = top->localDeclarations().at(0);
+    QCOMPARE(a->identifier().toString(), QString("a"));
+    compareUses(a, QList<RangeInRevision>() << RangeInRevision(0, 52, 0, 53));
+
+    Declaration *b = top->localDeclarations().at(1);
+    QCOMPARE(b->identifier().toString(), QString("b"));
+    compareUses(b, QList<RangeInRevision>() << RangeInRevision(0, 54, 0, 55));
+
+    Declaration *c = top->localDeclarations().at(2);
+    QCOMPARE(c->identifier().toString(), QString("c"));
+    compareUses(c, QList<RangeInRevision>() << RangeInRevision(0, 56, 0, 57));
+}
+
+void TestUses::closureParameterUnionType() {
+    //                 0         1         2         3         4         5         6
+    //                 0123456789012345678901234567890123456789012345678901234567890
+    QByteArray method("<? class A {} class B {} class C {} $a = function (A|B|C $a) {};");
+    TopDUContext *top = parse(method, DumpNone);
+    DUChainReleaser releaseTop(top);
+    DUChainWriteLocker lock(DUChain::lock());
+
+    Declaration *a = top->localDeclarations().at(0);
+    QCOMPARE(a->identifier().toString(), QString("a"));
+    compareUses(a, QList<RangeInRevision>() << RangeInRevision(0, 51, 0, 52));
+
+    Declaration *b = top->localDeclarations().at(1);
+    QCOMPARE(b->identifier().toString(), QString("b"));
+    compareUses(b, QList<RangeInRevision>() << RangeInRevision(0, 53, 0, 54));
+
+    Declaration *c = top->localDeclarations().at(2);
+    QCOMPARE(c->identifier().toString(), QString("c"));
+    compareUses(c, QList<RangeInRevision>() << RangeInRevision(0, 55, 0, 56));
+}
+
+void TestUses::closureReturnUnionType() {
+    //                 0         1         2         3         4         5         6
+    //                 0123456789012345678901234567890123456789012345678901234567890
+    QByteArray method("<? class A {} class B {} class C {} $a = function (): A|B|C {};");
+    TopDUContext *top = parse(method, DumpNone);
+    DUChainReleaser releaseTop(top);
+    DUChainWriteLocker lock(DUChain::lock());
+
+    Declaration *a = top->localDeclarations().at(0);
+    QCOMPARE(a->identifier().toString(), QString("a"));
+    compareUses(a, QList<RangeInRevision>() << RangeInRevision(0, 54, 0, 55));
+
+    Declaration *b = top->localDeclarations().at(1);
+    QCOMPARE(b->identifier().toString(), QString("b"));
+    compareUses(b, QList<RangeInRevision>() << RangeInRevision(0, 56, 0, 57));
+
+    Declaration *c = top->localDeclarations().at(2);
+    QCOMPARE(c->identifier().toString(), QString("c"));
+    compareUses(c, QList<RangeInRevision>() << RangeInRevision(0, 58, 0, 59));
+}
+
+void TestUses::propertyUnionType() {
+    //                 0         1         2         3         4         5         6
+    //                 01234567890123456789012345678901234567890123456789012345678901234567
+    QByteArray method("<? class A {} class B {} class C {} class D { public A|B|C $foo; }");
+    TopDUContext *top = parse(method, DumpNone);
+    DUChainReleaser releaseTop(top);
+    DUChainWriteLocker lock(DUChain::lock());
+
+    Declaration *a = top->localDeclarations().at(0);
+    QCOMPARE(a->identifier().toString(), QString("a"));
+    compareUses(a, QList<RangeInRevision>() << RangeInRevision(0, 53, 0, 54));
+
+    Declaration *b = top->localDeclarations().at(1);
+    QCOMPARE(b->identifier().toString(), QString("b"));
+    compareUses(b, QList<RangeInRevision>() << RangeInRevision(0, 55, 0, 56));
+
+    Declaration *c = top->localDeclarations().at(2);
+    QCOMPARE(c->identifier().toString(), QString("c"));
+    compareUses(c, QList<RangeInRevision>() << RangeInRevision(0, 57, 0, 58));
+}
+
+
+}
+
+#include "moc_uses.cpp"
